-- Copyright 1995-2000 Tom Moran, may be used by anyone for any purpose.

with Ada.Numerics,
     Ada.Numerics.Elementary_Functions,
     Ada.Unchecked_Deallocation,
     Claw.Bitmaps,
     Claw.Brushes,
     Claw.Rectangles,
     Claw.Window_Operations;

package body Display is

  Blue    : Claw.Colors.Color_Type := Claw.Colors.Rgb(0, 0, 255);
  Red     : Claw.Colors.Color_Type := Claw.Colors.Rgb(255, 0, 0);
  Black   : Claw.Colors.Color_Type := Claw.Colors.Rgb(0, 0, 0);
  Green   : Claw.Colors.Color_Type := Claw.Colors.Rgb(0, 255, 0);
  -- Note: the actual colors in the current palette may differ somewhat.

  type Eye_Positions is record
    X,Y : Float;
  end record;

  Eye_Color: constant array (Eyes) of Claw.Colors.Color_Type
    := (Left => Blue, Right => Red);

  Background_Brush: Claw.Brushes.Brush_Type
    := Claw.Brushes.Create(Claw.Brushes.Black);

  function Mouse_Has_Moved(Window : Cloud_Display_Type) return Boolean is
  begin
    return Window.Mouse_Moved;
  end Mouse_Has_Moved;

  procedure Open(Window    : in out Cloud_Display_Type;
                 Title     : in     String;
                 Annotation: in     XY_Pair_Lists := Null_Annotation;
                 Stereo    : in     Boolean := False;
                 Position  : in     Claw.Rectangle_Type := Claw.Basic_Window.Use_Default;
                 Parent    : in     Claw.Any_Window_Access_Type := null) is
  -- If Parent is null, the window will be a normal window with a border
  -- allowing movement, min/maximizing, or closing.  If Parent is not null,
  -- the window will be opened as a child window without a border.  If
  -- Parent is not null, but Position is the default, then the window's
  -- initial position will be the client area of the parent.

    use type Claw.Any_Window_Access_Type;
    use type Claw.Rectangle_Type;
    use type Claw.Int;
    New_Point_Handler: New_Point_Gateway_Access_Type;
  begin
    if Is_Valid(Window) then
      raise Claw.Already_Valid_Error;
    end if;
    Window.Is_Initialized := False;
    New_Point_Handler := new New_Point_Gateway(Window.Max_Particle_Count);
    Window.Is_Closed := False;
    if Stereo then
      Window.First_Eye := Eyes'First;
    else
      Window.First_Eye := Eyes'Last;
    end if;
    Window.Mouse_Moved := False;
    Window.Prev_Button := False;
    Window.Mouse_X := 0;
    Window.Mouse_Y := 0;
    Window.Next_Memory := Memory_Indices'First;
    Window.New_Point_Handler := New_Point_Handler;
    Window.LR := 0;
    Window.UD := 0;
    if Annotation'Length > 0 then
      Window.XY_Pair_List := new XY_Pair_Lists'(Annotation);
    end if;
    Window.Annotation_Present := False;
    Window.Memory := (others => (Null_Point, Null_Eye_Projection, False));
    if Parent /= null then
      Create(Window => Window, Window_Name => Title,
             Style => Claw.Styles.Child,
             Position => Claw.Client_Area(Parent.all),
             Parent => Parent.all);
    else
      Create(Window => Window, Window_Name => Title,
             Position => Position);
    end if;

    Show(Window => Window, How => Claw.Codes.Show_Startup);

    Set_Background_Brush(Window, Background_Brush);

    Window.Is_Initialized := True;

    Full_Redraw(Window);
    delay 0.1; -- give it a moment to paint with background

  end Open;

  procedure Close(Window : in out Cloud_Display_Type) is
  begin
    if Window.Is_Closed then
      return;
    end if;
    Destroy(Window);
  exception
    when Claw.Not_Valid_Error =>
      null;
  end Close;

  protected body New_Point_Gateway is

    procedure Put(List : in Point_Lists) is
    -- append List, wrapping around if necessary
    begin
      for I in List'range loop
        New_Point_List(di) := List(I);
        if Count < New_Point_List'Length then
          Count := Count + 1;
        end if;
        if di = New_Point_List'Last then
          di := New_Point_List'First;
        else
          di := di + 1;
        end if;
      end loop;
    end Put;

    procedure Get(List : out Point_Lists;
                  Last : out Integer) is
    -- return the (possibly empty) new_point_list
    -- leaving new_point_list empty
      N : constant Integer := Count;
    begin
      Count := 0;
      di := 1;
      List(1 .. N) := New_Point_List(1 .. N);
      Last := N;
    end Get;

  end New_Point_Gateway;

  procedure Small_Redraw(Window : in out Cloud_Display_Type) is
    use type Claw.Int;
  begin
    Claw.Window_Operations.Invalidate(Window, Erase_Background=>False);
  end Small_Redraw;

  procedure Full_Redraw(Window : in out Cloud_Display_Type) is
  begin
    Claw.Window_Operations.Invalidate(Window, Erase_Background=>True);
    delay 0.1;
  end Full_Redraw;


  procedure When_Draw(Window  : in out Cloud_Display_Type;
                      Easel   : in out Claw.Canvas.Basic_Canvas_Type'Class;
                      Erase_Background: in Boolean;
                      Area    : in     Claw.Rectangle_Type) is
    use type Claw.Int;
    use type Claw.Size_Type;
    Window_Size: constant Claw.Size_Type
      := Claw.Window_Operations.Client_Size(Window);
    X_Center: constant Float := Float(Window_Size.Width) / 2.0;
    Y_Center: constant Float := Float(Window_Size.Height) / 2.0;
    Window_Scale: constant Float -- don't quite fill window
      := 0.99*Float(Claw.Int'Min((Window_Size.Width*3)/4, -- CRT 4/3 aspect
                                  Window_Size.Height));

    Eye_Position: array (Eyes) of Eye_Positions := (Left  => (0.47, 0.5),
                                                    Right => (0.53, 0.5));
    -- moved together if monocular

    function Project(X,Y,Z : Unit_Cube_Locations) return Eye_Projections is
    --    pragma all_checks(off);
      Eye_Projection: Eye_Projections;
      Px,Py : Float;

      D : constant Float := 1.5;
      -- z distance from eyes to projection plane

      Projector: constant Float := D / (Z + D);

    begin
      for Eye in Window.First_Eye .. Eyes'Last loop
        Px := (X - Eye_Position(Eye).X) * Projector + 0.5;
        Py := (Y - Eye_Position(Eye).Y) * Projector + 0.5;

        if Px in 0.0 .. 1.0 and Py in 0.0 .. 1.0 then
          Eye_Projection(Eye)
            := (X => Claw.Int((Px - 0.5) * Window_Scale + X_Center),
                Y => Claw.Int((Py - 0.5) * Window_Scale + Y_Center));
        else
          Eye_Projection(Eye) := Null_Eye_Projection(Eye);
        end if;
      end loop;
      return Eye_Projection;
      --    pragma all_checks(on);
    end Project;

    procedure Paint(X     : in Claw.Int;
                    Y     : in Claw.Int;
                    Color : in Claw.Colors.Color_Type) is
    begin
      Claw.Bitmaps.Set_Pixel(Easel, (X, Y), Color,
                             Claw.Bitmaps.Ignore_all_Device_Errors);
    end Paint;

    Memory  : Memory_Array_Type renames Window.Memory;

    New_Point_Buffer: Point_Lists(1 .. Window.Max_Particle_Count);
    New_Point_Count: Integer;

    use type Claw.Rectangle_Type;

  begin
    if not Window.Is_Initialized then return;end if;

    if Window_Size = (0,0) then return;end if;

    if Window.First_Eye = Eyes'Last then -- view monocularly
      Eye_Position(Eyes'last).X
        := (Eye_Position(Eyes'First).X+Eye_Position(Eyes'Last).X)/2.0;
    end if;
    -- handle annotation
    Window.Annotation_Present
      := Window.Annotation_Present and not Erase_Background;
    if Window.Annotation_Present
    or Window.XY_Pair_List = null or Window.Mouse_Moved then
      null;
    else
      for I in Window.XY_Pair_List.all'range loop
        begin
          Paint(Claw.Int((Window.XY_Pair_List.all(I).X-0.5)*Window_Scale+X_Center),
                Claw.Int((Window.XY_Pair_List.all(I).Y-0.5)*Window_Scale+Y_Center),
                Green);
        exception
          when others =>
            null;
        end;
      end loop;
      Window.Annotation_Present := True;
    end if;
    -- end handle annotation

    -- get any new points
    Window.New_Point_Handler.Get(New_Point_Buffer, New_Point_Count);

    -- handle erasing old points
    declare
      Count : Integer := New_Point_Count;
      di : Memory_Indices := Window.Next_Memory;
    begin
      -- if mouse moving, erase all
      if Window.Mouse_Moved then Count := Memory'length;end if;
      -- if window size has changed, all are already erased
      if Erase_Background then
        Count := 0;
        for J in Memory'range loop
          Memory(J).Projection_Is_Valid := False;
        end loop;
      end if;
      for J in 1 .. Count loop
        if Memory(di).Projection_Is_Valid then
          for Eye in Window.First_Eye .. Eyes'Last loop
            Paint(Memory(di).Eye_Projection(Eye).X,
                  Memory(di).Eye_Projection(Eye).Y,
                  Black);
          end loop;
          Memory(di).Projection_Is_Valid := False;
        end if;
        if di = Memory'Last then
          di := Memory'First;
        else
          di := di + 1;
        end if;
      end loop;
    end; -- handle erasing old points

    -- handle fetch and save away new points
    declare
      di : Memory_Indices := Window.Next_Memory;
    begin
      for J in 1 .. New_Point_Count loop
        Memory(di).Point := New_Point_Buffer(J);
        if di = Memory'Last then
          di := Memory'First;
        else
          di := di + 1;
        end if;
      end loop;
    end; -- handle fetch and save away new points

    -- handle paint new points
    declare
      use Ada.Numerics, Ada.Numerics.Elementary_Functions;
      Count : Integer := New_Point_Count;
      di : Memory_Indices := Window.Next_Memory;
      Theta : constant Float := float(Window.UD) * pi/180.0;
      Phi   : constant Float := float(Window.LR) * pi/180.0;
      Cos_Theta : constant Float := Cos(Theta);
      Sin_Theta : constant Float := Sin(Theta);
      Cos_Phi   : constant Float := Cos(Phi);
      Sin_Phi   : constant Float := Sin(Phi);
      X, Y, Z : Float;
    begin
      if Window.Mouse_Moved or Erase_Background then
        Count := Memory'length;
      end if;
      for J in 1 .. Count loop
        if Memory(di).Point /= Null_Point then
          if not Window.Mouse_Moved then
            X := Window.Memory(di).Point.X;
            Y := Window.Memory(di).Point.Y;
            Z := Window.Memory(di).Point.Z;
          else
            Y := Cos_Theta * (Window.Memory(di).Point.Y - 0.5)
                 + Sin_Theta * (Window.Memory(di).Point.Z - 0.5)
                 + 0.5;
            Z := - Sin_Theta * (Window.Memory(di).Point.Y - 0.5)
                 + Cos_Theta * (Window.Memory(di).Point.Z - 0.5)
                 + 0.5;
            X := Cos_Phi * (Window.Memory(di).Point.X - 0.5)
                 + Sin_Phi * (Z - 0.5)
                 + 0.5;
            Z := - Sin_Phi * (Window.Memory(di).Point.X - 0.5)
                 + Cos_Phi * (Z - 0.5)
                 + 0.5;
          end if;
          if X in Unit_Cube_Locations
          and Y in Unit_Cube_Locations
          and Z in Unit_Cube_Locations then
            Memory(di).Eye_Projection := Project(X, Y, Z);
            Memory(di).Projection_Is_Valid := True;
            for Eye in Window.First_Eye .. Eyes'Last loop
              Paint(Memory(di).Eye_Projection(Eye).X,
                    Memory(di).Eye_Projection(Eye).Y,
                    Eye_Color(Eye));
            end loop;
          end if;
        end if;
        if di = Memory'Last then
          di := Memory'First;
        else
          di := di + 1;
        end if;
      end loop;
    end; -- handle paint new points
    Window.Next_Memory := Window.Next_Memory + New_Point_Count;
    if Window.Next_Memory > Memory'last then
      Window.Next_Memory := Window.Next_Memory - Window.Max_Particle_Count;
    end if;
  exception
    when Oops : others =>
      null; -- just forget it
  end When_Draw;

  procedure When_Mousemove(Window  : in out Cloud_Display_Type;
                           Param   : in     Claw.Mouse.Mouse_Params_Type;
                           Do_Default: in out Boolean) is
    use type Claw.Short;

    procedure Move_Eye_Position(LR,UD : in Integer) is
      -- adjust eye by Left-Right, Up-Down amounts
    begin
      Window.LR := (Window.LR + LR) mod 360;
      Window.UD := (Window.UD + UD) mod 360;
    end Move_Eye_Position;

  begin
    if not Param.Left_Button then  -- just ignore mouse with left button up
      Window.Prev_Button := False;
      return;
    end if;
    if not Window.Prev_Button then -- button has just been pushed, setup.
      Window.Mouse_X := Param.X;
      Window.Mouse_Y := Param.Y;
      Window.Prev_Button := True;
      return;
    end if;
    -- Left button remains pushed.  Use mouse delta.
    Do_Default := False;
    if abs (Param.X - Window.Mouse_X) > 50
    or abs (Param.Y - Window.Mouse_Y) > 50
      then
      -- a really big move - treat as a reset for mouse base
      Window.Mouse_X := Param.X;
      Window.Mouse_Y := Param.Y;
    end if;
    Move_Eye_Position(LR => Integer(Param.X - Window.Mouse_X),
                      UD => Integer(Param.Y - Window.Mouse_Y));
    Window.Mouse_X := Param.X;
    Window.Mouse_Y := Param.Y;
    if Window.Mouse_Moved then
      Small_Redraw(Window);
    else
      Window.Mouse_Moved := True;
      Full_Redraw(Window);  -- first mouse movement, erase annotation
    end if;
  end When_Mousemove;

  function Is_Closed(Window : Cloud_Display_Type) return Boolean is
  -- return True after window is closed
  begin
    return Window.Is_Closed;
  end Is_Closed;

  procedure Brighten(Window : in out Cloud_Display_Type;
                     X,Y,Z  : in     Unit_Cube_Locations) is
  begin
    Brighten(Window, (1 => (X, Y, Z)));
  end Brighten;

  procedure Brighten(Window : in out Cloud_Display_Type;
                     List   : in     Point_Lists) is
  begin
    if Window.Is_Closed then return;end if;
    Window.New_Point_Handler.Put(List);
    Small_Redraw(Window);
    Claw.Yield;
  exception
    when Claw.Not_Valid_Error =>
      null;
  end Brighten;

  procedure When_Close(Window : in out Cloud_Display_Type) is
  begin
    Window.Is_Closed := True;
    Claw.Basic_Window.When_Close(Claw.Basic_Window.Basic_Window_Type(Window));
  end When_Close;

  procedure Free is new Ada.Unchecked_Deallocation
    (New_Point_Gateway, New_Point_Gateway_Access_Type);

  procedure Free is new Ada.Unchecked_Deallocation
    (XY_Pair_Lists, XY_Pair_List_Pointers);

  procedure Finalize(Window : in out Cloud_Display_Type) is
  begin
    Free(Window.New_Point_Handler);
    Free(Window.XY_Pair_List);
    Claw.Basic_Window.Finalize(Claw.Basic_Window.Basic_Window_Type(Window));
  end Finalize;

end Display;
