-- This component displays a "cloud" of a fixed number of points in the
-- three dimensional unit cube.  They can be projected on the screen
-- either in 2 dimensions, or, with blue-red stereo glasses, in "3D".
-- Their orientation can be rotated by mouse movement.  Once the full set
-- of points has been set, new points replace older ones.  If the new
-- points are slightly offset from the points they replace, the set
-- appears to move smoothly.  If the new points are highly offset, and in
-- different ways, the effect is more of a "sparkle".  If there are no new
-- points after the initial set, the shape is static except as rotated by
-- mouse movement.

-- Copyright 1995-2000 Tom Moran, may be used by anyone for any purpose.

with Claw,
     Claw.Basic_Window,
     Claw.Canvas;

package Display is

  type Cloud_Display_Type(Max_Particle_Count: Positive)
  is new Claw.Basic_Window.Basic_Window_Type with private;

  subtype Unit_Cube_Locations is Float range 0.0 .. 1.0;

  type XY_Pairs is record
  -- for annotations
    X,Y : Unit_Cube_Locations; -- X increases left to right, Y top to bottom
  end record;

  -- The set of XY_Pairs are intended for annotation.  They show
  -- up in green and disappear at the first movement of the mouse.

  type XY_Pair_Lists is array (Integer range <>) of XY_Pairs;

  Null_Annotation: XY_Pair_Lists(1 .. 0);

  procedure Open(Window    : in out Cloud_Display_Type;
                 Title     : in     String;
                 Annotation: in     XY_Pair_Lists := Null_Annotation;
                 Stereo    : in     Boolean := False;
                 Position  : in     Claw.Rectangle_Type := Claw.Basic_Window.Use_Default;
                 Parent    : in     Claw.Any_Window_Access_Type := null);
  -- If Parent is null, the window will be a normal window with a border
  -- allowing movement, min/maximizing, or closing.  If Parent is not null,
  -- the window will be opened as a child window without a border.  If
  -- Parent is not null, but Position is the default, then the window's
  -- initial position will be the client area of the parent.

  procedure Small_Redraw(Window : in out Cloud_Display_Type);
  -- just draw new points

  procedure Full_Redraw(Window : in out Cloud_Display_Type);
  -- erase and redraw

  procedure Close(Window  : in out Cloud_Display_Type);

  type Points is record    -- X increases left to right, Y top to bottom
    X,Y,Z : Unit_Cube_Locations; -- Z increases away from observer.
  end record;

  -- The Point_Lists are the changing points.  If the window was
  -- opened in Stereo, they will be projected as blue-red pairs,
  -- otherwise they will be shown as single red points.

  type Point_Lists is array (Integer range <>) of Points;

  procedure Brighten(Window : in out Cloud_Display_Type;
                     X,Y,Z  : in     Unit_Cube_Locations);

  procedure Brighten(Window : in out Cloud_Display_Type;
                     List   : in     Point_Lists);

  function Mouse_Has_Moved(Window : Cloud_Display_Type) return Boolean;
  -- becomes True when Mouse moves

  function Is_Closed(Window : Cloud_Display_Type) return Boolean;
    -- return True after window is closed

private

  type XY_Pair_List_Pointers is access XY_Pair_Lists;

  subtype Memory_Indices is Positive;

  Null_Point: constant Points := (0.0, 0.0, 0.0);

  type Eyes is (Left, Right);

  subtype Screen_Points is Claw.Point_Type;

  -- Two projection as seen by the two eyes (stereo) or just Right (2D)
  type Eye_Projections is array (Eyes) of Screen_Points;

  Null_Eye_Projection: constant Eye_Projections := (Eyes => (0, 0));

  type Memory_Type is record
    Point   : Points := Null_Point;
    Eye_Projection: Eye_Projections := Null_Eye_Projection;
    Projection_Is_Valid: Boolean := False;
  end record;

  type Memory_Array_Type is array (Memory_Indices range <>) of Memory_Type;

  protected type New_Point_Gateway(Max_Count: Positive) is
    procedure Put(List : in     Point_Lists);
    procedure Get(List :    out Point_Lists;
                  Last :    out Integer);
  private
    New_Point_List: Point_Lists(1 .. Max_Count);
    Count   : Natural := 0;
    di      : Integer := 1;
  end New_Point_Gateway;

  type New_Point_Gateway_Access_Type is access New_Point_Gateway;

  type Cloud_Display_Type(Max_Particle_Count: Positive)
  is new Claw.Basic_Window.Basic_Window_Type with record
    Is_Initialized : Boolean := False; -- don't When_Draw till True
    Is_Closed: Boolean := True;
    First_Eye: Eyes := Eyes'First; -- paint (first_eye .. eyes'last)
    -- so first_eye=eyes'first means stereo, first_eye=eyes'last means mono
    Mouse_Moved: Boolean := False;
    Prev_Button: Boolean := False;
    Mouse_X,
    Mouse_Y : Claw.Short;
    Next_Memory: Memory_Indices := Memory_Indices'First;
    LR, UD : Integer := 0;  -- left-right, up-down mouse movement
    New_Point_Handler: New_Point_Gateway_Access_Type(Max_Particle_Count);
    Annotation_Present : Boolean := False;
    XY_Pair_List: XY_Pair_List_Pointers;
    Memory  : Memory_Array_Type(1 .. Max_Particle_Count);
  end record;

  -- When_Draw et al are in the package spec for dispatching

  procedure When_Draw(Window  : in out Cloud_Display_Type;
                      Easel   : in out Claw.Canvas.Basic_Canvas_Type'Class;
                      Erase_Background: in Boolean;
                      Area    : in     Claw.Rectangle_Type);

  procedure When_Mousemove(Window  : in out Cloud_Display_Type;
                           Param   : in     Claw.Mouse.Mouse_Params_Type;
                           Do_Default: in out Boolean);

  procedure When_Close(Window  : in out Cloud_Display_Type);

  procedure Finalize(Window : in out Cloud_Display_Type);

end Display;
